#
# Copyright (C) 2020 IHS Markit.
# All Rights Reserved
#

class dataset_urls:
    datafiles = '/__api/datasets/{id}/datafiles/'
    latest_datafile = '/__api/datasets/{id}/latest-datafile/'

    # This endpoint returns either RO or RW dependent on your permissions
    # and this depends on the role you have on the package of the dataset you are requesting it for
    # User is Data Steward + Technical Support for Package parent Dataset = RW (read/write from/to s3)
    # User is not = RO (read from s3)

    # As a user, you need this in the SDK if you are trying to access the files directly. Catalogue takes care of the
    # RO/RW permissioning, so if you can access the dataset, you can access the data. If you use S3-Proxy, you use
    # the S3-Proxy's permissions (via your JWT), whereas here, you get the keys and then can use them.
    access_keys = '/__api/request-access-keys/'

    # v2
    v2_index = '/__api_v2/datasets/'
    v2_by_id = '/__api_v2/datasets/{id}/'
    v2_by_short_code = '/__api_v2/by_short_code/dataset/{dataset_short_code}'
    v2_sample_data_schema = '/__api_v2/datasets/{id}/sample_data'
    v2_sample_data_file = '/__api_v2/datasets/{id}/sample_data/file'
    v2_schema_instance_version = '/__api_v2/datasets/{id}/dictionaries/{version}'
    v2_unstructured_document = '/__api_v2/datasets/{id}/documents'

    dictionary_index =  '/__api_v2/dictionaries/'
    dictionary_instance = '/__api_v2/dictionaries/{id}/'
    dictionary_by_dataset = '/__api_v2/datasets/{id}/dictionaries/'
    dictionary_by_dataset_lastest = '/__api_v2/datasets/{id}/current_dictionary/'
    dictionary_fields = '/__api_v2/dictionaries/{id}/fields/'


class autoreg_urls:
    autoreg_index = '/__api/auto-reg-metadata/'
    autoreg_instance = '/__api/auto-reg-metadata/{id}/'


class datafile_urls:
    datafiles_index = '/__api/datafiles/'
    datafiles_instance = '/__api/datafiles/{id}/'


class package_urls:

    # v2
    v2_package_datasets = '/__api_v2/packages/{id}/datasets'
    v2_package_by_id = '/__api_v2/packages/{id}'
    v2_package_index = '/__api_v2/packages'


class consumption_urls:
    consumption_download = '/datafile/{id}/download/binary/{path}'
    consumption_manifest = '/datafile/{id}/manifest/'
    consumption_dataframe = '/dataset/{id}/dataframe/'
    consumption_analytics = '/analytics/'
    consumption_partitions = '/dataset/{id}/partitions/'


class sam_urls:
    sam_token = '/sso/oauth2/realms/root/realms/Customers/access_token'


class sam_api_urls:
    sam_pat_token = '/authentication/v1/Authenticate/PersonalAccessToken'


class identity_urls:
    identity_token = '/api/identity/v2/auth/token'
    identity_postbox = '/api/identity/v1/postbox_login'
    identity_poll = '/api/identity/v2/auth/postbox'
    orgs_visible_to_user = '/api/identity/v2/organisations/visible'


#
# Copyright (C) 2019 IHS Markit.
# All Rights Reserved
#

import io
import logging
from pyarrow.lib import Buffer


class DequeueFile(io.IOBase):
    """
    Fake interface to output a arrow IPC stream. A pyarrow
    RecordBatchStreamWriter can only take a file like object.
    Since Flask can't return that without buffering the enitre
    response, which can be Terabytes! we create a fake file object
    that doesn't buffer.

    N.B inherits from io.BytesIO because pyarrow does a typecheck
    on files passed in as a sink.
    """

    logger = logging.getLogger(__name__)

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        # TODO investigate using the collections.deque datastructure
        self.queue = []

    def __del__(self):
        """Destructor."""
        self.logger.debug('Destructor called on DequeueFile')

    def write(self, data):
        # TODO ask Artur why this wasn't a problem in testing.
        # sometimes pyarrow seems to be passing in a buffer
        # rather than a byte string. There's possibly a way
        # of steaming this raw buffer without doing a copy, i.e
        # improving how the dequeue function works.
        if type(data) == Buffer:
            self.queue.append(data.to_pybytes())
        else:
            self.queue.append(data)

    def writable(self):
        return True

    def dequeue(self):
        if len(self.queue) == 0:
            self.logger.warning('dequeue returning an empty array')

        yield from self.queue

        # Explicitly clear the list.
        self.queue.clear()



#
# Copyright (C) 2019 IHS Markit.
# All Rights Reserved
#
import io
import logging
import dataclasses
import itertools
import os
import pandas
import pandas.io.parsers
import pyarrow
import numpy
import re
from typing import Iterator, Pattern, Optional, Dict
from pyarrow import RecordBatchStreamWriter, RecordBatch

from ..handlers.dequeue_file import DequeueFile
from ...arrow_service.exceptions import (NoFilesInDataset, NonCsvFilesInDataset)
from ....services.exceptions import ServiceException, SchemaMismatchException
from ....adapters.s3_io_adapters import ServiceIO
from ....services.arrow_service.handlers.stream_errors import (handle_arrow_stream_errors)


logger = logging.getLogger(__name__)


@dataclasses.dataclass
class Config:
    # Chunk size is the number of rows read per chunk.
    # TODO scott: what is the optimal chunk size? It's hard to guess
    # as each .csv file can have a different number of columns.
    # For Storm data:
    # chunk size 1000 body buffer size 2097152 time > 4 minutes, memory usage
    # is low.
    # chunk size 10000 body buffer size 2097152 time 0:03:04.3, 0:03:34.2,
    # memory usage <= 227 MB.
    # chunk size 10000 body buffer size 12582912 time 0:02:52.0, 0:02:57.9,
    # memory usage <= 232 MB
    # chunk size 100000 body buffer size 2097152 time 0:02:51.3, 0:02:53.8,
    # memory usage > 500 MB!
    chunk_size: int = int(os.environ.get('CSV_CHUNK_SIZE', 5000))
    header_buffer_size: int = int(
        os.environ.get('CSV_HEADER_BUFFER_SIZE', 1024)
    )
    body_buffer_size: int = int(
        os.environ.get('CSV_BODY_BUFFER_SIZE', 2097152)
    )


class CsvHandler:

    __file_extension_regex: Pattern[str] = re.compile(r'^.+\.(\w+)$')

    __pandas_supported_compression: Dict[str, str] = {
        # Note: The gzip file extension '.gz' is inconsistently named 'gzip'
        # in the pandas compression parameters, so we have to convert the name.
        'gz': 'gzip',
        'bz2': 'bz2',
        'zip': 'zip',
        'xz': 'xz',
    }

    @staticmethod
    def _extract_file_extension(path: str) -> Optional[str]:
        match = CsvHandler.__file_extension_regex.search(path)
        return match.group(1) if match else None

    def __init__(self, files: Iterator[ServiceIO],
                 order_by_latest: Optional[bool] = None, **kwargs):
        self.config = Config(**kwargs)
        self.order_by_latest = order_by_latest

        logger.debug('CSV handler constructor')

        initial_file = None

        if order_by_latest is True:
            # The parquet schema handling depends on us knowing which
            # is the latest .parquet file. The list of files has to
            # be ordered in some way (ascending or descending),
            # it cannot be un-ordered.
            logger.debug(
                'Use the schema from the first file in the list.'
            )
        else:
            # Use the schema from the first file in the list. This
            # will trigger a listing of the files so will be slower
            # to get the first bytes of data back to the client.

            # For the case of order_by_latest None, the default of S3
            # seems to be to return sorted ascending alphabetically,
            # ASSUMPTION this may be close enough. If not, we will have
            # to change the default in dataframe/views.py from None to
            # either True or False.
            logger.debug(
                'Use the schema from the last file in the list.'
            )

            # TODO for the additive schema we need to know the final
            #  file and will use that as the schema for the whole
            #  response. We need to cast to type List to action the
            #  iterator so we can flip the list. Alternative to
            #  caching the whole list is to instead cache only the
            #  S3File of the last file and process that file
            #  immediately (so nginx is able to return quickly),
            #  reading the rest of the files straight from the S3
            #  list not from cache and therefore we can keep `files`
            #  as type iterator. This would also require a crawler to
            #  populate cache with the latest as sorting takes time.
            files = iter(reversed(list(files)))

        try:
            initial_file = next(files)
        except StopIteration:
            raise NoFilesInDataset(
                'No data in dataset. Unable to stream.'
            )

        extension = CsvHandler._extract_file_extension(initial_file.path)
        # If the extension is:
        # 1. just '.csv',
        # 2. or is an unsupported file type,
        # 3. or there is no extension
        # then pass None as the compression type so that read_csv handles the
        # file as a plain csv file.
        logger.debug(
            f"Extracted file extension '{extension}'",
            extra={
                'extension': extension,
                'path': initial_file.path,
            }
        )

        self.validate_file_extension(initial_file)

        compression = \
            CsvHandler.__pandas_supported_compression.get(str(extension))

        logger.debug(
            'File extension maps to pandas file compression type '
            f"'{compression}'",
            extra={
                'compression': compression,
                'path': initial_file.path,
            }
        )

        try:
            # The gzip code attempts to scan the header one character at a
            # time looking for the end of header marker. This results in over
            # 85 requests to S3. Instead use a buffer because it will request
            # large chunks instead of individual characters, the result will be
            # cached and returned directly on subsequent reads.
            buf = io.BufferedReader(
                initial_file,
                buffer_size=self.config.header_buffer_size
            )

            df: pandas.DataFrame = pandas.read_csv(
                # Pandas will perform automatic on-the-fly decompression
                # based on file extension. Pandas versions > 0.18.1 supports
                # gzip, bz2, zip and xz.
                buf,
                # We are telling pandas to ignore blank lines at the start of a
                # file (sometimes these are introduced into .csv when a data
                # scientist has performed a manual edit on a file).
                skip_blank_lines=True,
                # The header is on the first line of data in the file (after
                # skipping blank lines).
                header=0,
                # Only read the header row from the data file, then stop
                # reading as we only need the header.
                nrows=0,
                compression=compression,
            )

            logger.debug(
                'Header parsed by pandas',
                extra={
                    'extension': extension,
                    'compression': compression,
                    'path': initial_file.path,
                }
            )

            # Detach the underlying stream otherwise your BufferedIOBase
            # object closes the file when it goes out of scope.
            # https://stackoverflow.com/a/8631332
            buf.detach()
            # Remember to rewind the stream! The side-effect of Pandas
            # reading the first line from the stream is that it leaves the
            # pointer pointing to the end of the line. You must manually
            # reset to the start of the file after reading the header so that
            # subsequent calls to the same file will behave as expected.
            initial_file.seek(0)
            logger.debug('seek is back to the start of the file')

        except (OSError, UnicodeDecodeError) as e:
            message = 'Error for initial file. ' \
                      'Unable to stream.'
            logger.warning(
                message,
                exc_info=e,
                extra={'path': initial_file.path}
            )
            raise NonCsvFilesInDataset(
                'Non csv file found in dataset. Path: '
                f'{initial_file.path}'
            ) from e

        except Exception as e:
            message = 'The file is corrupt, ' \
                      'not a .csv like file or cannot be read. ' \
                      'Please report the problem to the datalake ' \
                      'platform team at DataLake-Support_L1@ihsmarkit.com. ' \
                      'Please include the full error message including ' \
                      'the request_id.'
            logger.exception(message)
            raise ServiceException(message) from e

        # For the type of the column to be string as types in a .csv may not
        # be consistent across files.
        fields = [
            (column_name, pyarrow.string())
            for column_name in list(df.columns)
        ]

        logger.debug(
            'Column names read from .csv header row (cast to string)',
            extra={'columns': fields}
        )

        self.schema: pyarrow.Schema = pyarrow.schema(fields).remove_metadata()
        logger.debug(
            'Parquet columns mapped to pyarrow schema names',
            extra={
                'schema_names': self.schema.names
            }
        )
        logger.debug(
            'Parquet columns mapped to pyarrow schema types',
            extra={
                'schema_names': self.schema.types
            }
        )

        self._initial_file = initial_file

        if order_by_latest is True:
            self._files = itertools.chain([initial_file], files)
        else:
            # The list had been flipped from ascending to descending,
            # so now flip it back. This is clunky and would be better if
            # we only cached the initial schema file.
            self._files = itertools.chain(
                reversed(list(files)),
                [initial_file]
            )

    @staticmethod
    def validate_file_extension(s3_file: ServiceIO):
        if '.csv' not in s3_file.path:
            logger.warning(
                'Non csv file found in dataset.',
                extra={
                    'path': s3_file.path,
                }
            )
            raise NonCsvFilesInDataset(
                f'Non csv file found in dataset. Path: {s3_file.path}'
            )

    @staticmethod
    def _validate_file_size(s3_file: ServiceIO):
        """File size must be greater than zero."""
        if hasattr(s3_file, 'size'):
            return s3_file.size > 0
        else:
            return False

    def _get_row_chunks_from_files(self) -> Iterator[pandas.DataFrame]:
        """Returns a lazy iterator which gets batches multiple files into
        chunks of N lines . """
        for file_number, file in enumerate(self._files):
            if not self._validate_file_size(file):
                # Skip zero size files.
                logger.warning(
                    'File has been skipped, CSV file size is 0 bytes.',
                    extra={'path': file.path}
                )
                continue
            self.validate_file_extension(file)

            logger.debug(
                f"Iterating over file number '{file_number}'",
                extra={
                    'path': file.path,
                    'file_number': file_number
                }
            )

            for idx, file_reader in enumerate(self._get_row_chunks_from_file(file)):
                for chunks in file_reader:
                    chunk = pyarrow.RecordBatch.from_pandas(
                        chunks, preserve_index=False
                    )

                    current_csv_schema = chunk.schema.remove_metadata()

                    # If the current schema is not a subset
                    # TODO scott: we should also be able to work with
                    # transposed columns. This will also be necessary
                    # for when it is additive but they have added a
                    # column in the middle instead of at the end!
                    is_schema_ok = (
                        current_csv_schema.names ==
                        self.schema.names[:len(current_csv_schema)]
                    )
                    if not is_schema_ok:
                        raise SchemaMismatchException(
                            'Tried to process a file with a schema that does '
                            'not match the schema of the latest. We '
                            'require that all .csv files have an additive '
                            'schema.'
                            "\n\nLatest file's schema:"
                            f"\n\n{self.schema.remove_metadata().names}"
                            "\n\nLatest file's path:"
                            f'\n\n{self._initial_file.path}'
                            f"\n\nCurrent file's schema:"
                            f"\n\n{current_csv_schema.names}"
                            "\n\nCurrent file's path:"
                            f'\n\n{file.path}'
                        )

                    unused_fields = list(self.schema)[len(current_csv_schema):]
                    columns = list(chunk.columns)

                    for field in unused_fields:
                        # TODO scott: This may need a change if we have
                        # transposed columns, as we would need to append
                        # in a specific order.
                        columns.append(
                            # Append and empty array of the correct type and
                            # size to fill in the missing column.
                            pyarrow.array(
                                numpy.empty(chunk.num_rows, dtype=object),
                                field.type
                            )
                        )

                    pandas_chunk = RecordBatch.from_arrays(
                        columns, schema=self.schema
                    ).to_pandas()

                    yield file, pandas_chunk
            # The side-effect of Pandas reading the first line from the
            # stream is that it leaves the pointer pointing to the end of the
            # line. You must manually reset to the start of the file after
            # reading the header so that subsequent calls to the same file
            # will behave as expected.
            logger.debug('Seek to start of the file...')
            file.seek(0)
            logger.debug('Seek completed')

    def _get_row_chunks_from_file(self, file: ServiceIO):
        """Returns a lazy iterator which gets batches a single file into
        chunks of N lines . """

        extension = CsvHandler._extract_file_extension(file.path)
        # If the extension is:
        # 1. just '.csv',
        # 2. or is an unsupported file type,
        # 3. or there is no extension
        # then pass None as the compression type so that read_csv handles the
        # file as a plain csv file.
        logger.debug(
            f"Extracted file extension '{extension}'",
            extra={
                'extension': extension,
                'path': file.path,
            }
        )

        compression = \
            CsvHandler.__pandas_supported_compression.get(str(extension))

        logger.debug(
            'File extension maps to pandas file compression type '
            f"'{compression}'",
            extra={
                'extension': extension,
                'compression': compression,
                'path': file.path,
            }
        )

        chunks: pandas.io.parsers.TextFileReader = pandas.read_csv(
            # Pandas will perform automatic on-the-fly decompression based on
            # file extension. Pandas versions > 0.18.1 supports gzip, bz2,
            # zip and xz.
            # The gzip code attempts to scan the header one character at a
            # time looking for the end of header marker. This results in over
            # 85 requests to S3. Instead use a buffer because it will request
            # large chunks instead of individual characters, the result will be
            # cached and returned directly on subsequent reads.
            io.BufferedReader(file, buffer_size=self.config.body_buffer_size),
            # Force all column types to be string. We should NOT allow pandas
            # to infer the type as there are cases where the first file we
            # read does not have a large enough sample of data and so infers
            # the wrong type e.g. the data it sees in column 0 in the first
            # file only contains integers so it infers integer, but the
            # second file contains strings (or another type like floats). The
            # process would fail when trying to combine the separate
            # DataFrames because of the different types.
            dtype=str,
            # We are telling pandas to ignore blank lines at the start of a
            # file (sometimes these are introduced into .csv when a data
            # scientist has performed a manual edit on a file).
            skip_blank_lines=True,
            # The header is on the first line of data in the file (after
            # skipping blank lines).
            header=0,
            # Allows us to iterate through a very large file lazily rather
            # than reading the entire file into memory. The file is chunked
            # into N rows.
            chunksize=self.config.chunk_size,
            compression=compression,
        )
        logger.debug(
            'pandas has chunked the file with chunk_size '
            f"'{self.config.chunk_size}'",
            extra={
                'extension': extension,
                'compression': compression,
                'path': file.path,
                'chunk_size': self.config.chunk_size,
            }
        )
        yield chunks

    def handle(self, nrows=None) -> Iterator[bytes]:
        chunks_per_file = self._get_row_chunks_from_files()

        sink = DequeueFile()
        rows_iterated = 0

        try:
            with handle_arrow_stream_errors(
                RecordBatchStreamWriter(sink, self.schema)
            ) as writer:
                for idx, (file, chunk) in enumerate(chunks_per_file):
                    rows, cols = chunk.shape

                    logger.debug(
                        f"Iterating over CSV chunk number '{idx}'",
                        extra={
                            'chunk_number': idx,
                            'shape': (rows, cols),
                            'rows_iterated': rows_iterated,
                        }
                    )

                    if nrows and (rows_iterated + rows) > nrows:
                        logger.debug('Truncating.')
                        truncated_chunk = chunk.head(nrows - rows_iterated)

                        batch = RecordBatch.from_pandas(
                            df=truncated_chunk, schema=self.schema
                        )

                        writer.write_batch(batch)
                        yield from sink.dequeue()
                        break
                    else:
                        logger.debug('Record batch from pandas')
                        batch = RecordBatch.from_pandas(
                            df=chunk, schema=self.schema
                        )

                        logger.debug('write batch')
                        writer.write_batch(batch)

                        logger.debug('sink dequeue')
                        yield from sink.dequeue()
                        rows_iterated = rows_iterated + rows

                logger.debug('Completed iteration.')
        finally:
            yield from sink.dequeue()



import inspect
from typing import Callable


def __call_extract_fn(application_extract_func, target, function, arguments, keyword_args, class_fields_to_include=None):
    metadata, obj = application_extract_func(
        target, function, arguments, keyword_args, class_fields_to_include)
    return metadata, obj


def __get_analytic_fns_object_from_target(target) -> Callable:

    # HACK2: Firstly you can't not do this check, with MagicMock or UnstructuredDatasetModel, since
    # any unknown variable is a DeferInCaseCall, i.e. always exists, even if it doesnt.
    # Secondly, you can't check for _client, since a Model HAS_A client, but doesnt not have the method callback
    # that aspects needs to call, so we have to redirect the method call (target_a), but keep the target
    if not hasattr(target, "_analytics_extract_metadata") \
            or not inspect.ismethod(target._analytics_extract_metadata):  # magicmock/deferincasecall yield True
        # TODO dirty module, must check reference trait/properties of target and choose _client for SDK
        target_a = target._client._analytics_extract_metadata
    else:
        target_a = target._analytics_extract_metadata

    return target_a


def _make_function_call(aspects, target, function, args, kwargs, class_fields_to_include):

    '''
     Explanation of who the hell is target.
     ====
     on SDK Mixins -
     see Component_Aspect_Wrapper, handled whereby target (the mixin) extends the client,
     so target IS client which HAS _analytics_handler, _analytics_extract_metadata
     ---
     on SDK Models -
     target is e.g FileModel/DatasetModel, so target HAS a client
     and a client HAS _analytics_handler, _analytics_extract_metadata

     Because of UnstructuredDatasetModel, it HAS _client, but also states that it HAS any function/property
     as this gets turned into a deferincasecall, so will pass any check. The way we can capture this is to look
     for whether _analytics_extract_metadata (any property of the target - the one with the aspects required methods)
     is a function (or NOT an object)
     ---
     on S3Proxy -
     similar to mixin case
     target IS Dispatcher which HAS _analytics_handler, _analytics_extract_metadata
     ---
     on Tests -
     target is a magicmock and/or target._client is a magicmock. _analytics_extract_metadata should return a SkipError
     to skip over analytics, if there is no real implementation to use.
    '''

    def inner():
        try:
            result = function(target, *args, **kwargs)
        except TypeError as t:
            # HACK1: its a pity to put this here, but we need to hijack TypeError for UnstructuredDatasetModel
            # we can't use hasattr, since the TypeError hijack will return something!
            # N.B. if you need to see the real cause, comment out the _type_exception_handler of the class.
            # TODO dirty module, must check whether target is UnstructuredModel for SDK
            if "_type_exception_handler" in dir(target):
                target._type_exception_handler(t, function, args, kwargs)
                return
            else:
                for aspect in aspects:
                    aspect.invoke_after_exception_aspects(obj, metadata, t)
                raise t

        return result

    try:
        application_extract_func = __get_analytic_fns_object_from_target(target)
        metadata, obj = __call_extract_fn(application_extract_func, target, function, args, kwargs, class_fields_to_include)
    except Exception as e:
        # just because extraction for analytics fails, doesnt mean we should fail the call undecorated. Still do it!
        metadata = {}
        obj = application_extract_func
        return inner()

    # if we're here, then we can run the function and wrap with analytics as the extract passed,
    # we did not hit exception
    try:
        for aspect in aspects:
            aspect.invoke_pre_call_aspects(obj, metadata)

        result = inner()

        for aspect in aspects:
            aspect.invoke_post_call_aspects(obj, metadata)
    except Exception as e:
        # If the object contains a `strict` boolean and it is set to
        # True, then print out the exception message and a full stack
        # trace.
        # DEFAULTED to True to match the current behaviour. This default
        # can be changed in a later release. When we change the default
        # we will have to update the tests to explicitly set strict to
        # True to maintain the previous test behaviour.
        if getattr(obj, 'strict', True):
            for aspect in aspects:
                aspect.invoke_after_exception_aspects(obj, metadata, e)
            raise e
        elif getattr(obj, 'logger', None):
            # Data scientists do not want to see stack dumps by
            # default, especially when we have a root cause that
            # triggers secondary exceptions.
            obj.logger.warning(
                '>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>'
                '\nAn exception occurred so we are returning None.'
                '\nTo see the exception and stack trace, please start '
                'the session again with the parameter `strict=True`'
                '\n<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<'
            )
            return None

    return result



"""Packaging settings."""
import subprocess
from setuptools import find_packages, setup


def get_git_revision_short_hash() -> str:
    try:
        return subprocess.check_output(['git', 'rev-parse', '--short', 'HEAD']).decode('ascii').strip()
    except:
        # if this is packaged up, then there is no git version (since its been packaged up and above will fail
        return "0.0.0"

try:
    with open('requirements.txt', 'r') as f:
        requirements = f.readlines()
except:
    # if this is packaged up, requirements.txt isn't present, its in the /tmp dir of the package
    with open('platform_services_lib.egg-info/requires.txt') as f:
        requirements = f.readlines()

setup(
    name='platform-services-lib',
    version=get_git_revision_short_hash(),
    python_requires='>3.6.0',
    url='https://ihsmarkit.com',
    description='Data Lake services lib.',
    author='IHS Markit Data Lake Team',
    author_email='datalake-support@ihsmarkit.com',
    license='MOZ-2',
    classifiers=[
        'Intended Audience :: Developers',
        'Topic :: Software Development :: Libraries :: Python Modules',
        'License :: OSI Approved :: Mozilla Public License 2.0 (MPL 2.0)',
        'Natural Language :: English',
        'Operating System :: OS Independent',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.6',
    ],
    keywords='datalake, data, lake',
    packages=(find_packages(exclude=['docs', 'tests'])),
    install_requires=requirements
)



#
# Copyright (C) 2019 IHS Markit.
# All Rights Reserved
#
import dataclasses
from typing import Optional


@dataclasses.dataclass
class AnalyticsEvent:
    application_name: str
    application_version: str
    user_id: str
    entity: str
    action: Optional[str] = None
    organisation_id: Optional[str] = None
    result: Optional[int] = None
    properties: Optional[dict] = None



import logging
import os
import uuid
import jwt
import requests
from urllib.parse import urljoin
from typing import Optional

from ..context.environments import _Environment
from ..context.urls import sam_urls, sam_api_urls, identity_urls
from ..providers.oidc_provider import JWTEncoded
from ..services.view_exceptions import Unauthorized

logger = logging.getLogger(__name__)


def _sam_app_flow(environment: _Environment, user_id: str, user_secret: str, origin=None) -> Optional[str]:

    if user_id is None:
        logger.warning("SAM App User ID is not set")

    if user_secret is None:
        logger.warning("SAM App User Secret is not set")

    if environment.sam is None:
        logger.warning("SAM Endpoint is not set")

    if environment.catalogue is None:
        logger.warning("Catalogue Endpoint is not set")

    catalogue_url = urljoin(environment.catalogue, identity_urls.identity_token)
    sam_url = urljoin(environment.sam, sam_urls.sam_token)

    # exchange the client credentials with SAM for an access token
    sam_payload = {
        'client_id': user_id,
        'client_secret': user_secret,
        'grant_type': 'client_credentials'
    }

    logger.info("Exchanging details with SAM")
    resp = external_session().post(sam_url, data=sam_payload)
    if resp.status_code != 200:
        raise Unauthorized('Could not retrieve jwt access token with client credentials')

    # exchange the acquired token with catalogue
    sam_token = resp.json()['access_token']
    dl_payload = {
        'client_id': user_id,
        'subject_token': sam_token
    }

    if origin:
        # this allows influencing JWT lifetime (1hr -> 8hr)
        dl_payload["origin"] = origin

    logger.info("Exchanging SAM token with Catalogue")
    resp = external_session().post(catalogue_url, data=dl_payload)
    if resp.status_code != 200:
        raise Unauthorized('Could not retrieve jwt access token with client credentials')

    return resp.json()['access_token']


def external_session():
    return requests.Session()


def _sam_pat_flow(environment: _Environment, user_id: str, user_secret: str) -> Optional[str]:

    if user_id is None:
        logger.warning("SAM PAT User ID is not set")

    if user_secret is None:
        logger.warning("SAM PAT User Secret is not set")

    if environment.sam is None:
        logger.warning("SAM Endpoint is not set")

    if environment.catalogue is None:
        logger.warning("Catalogue Endpoint is not set")

    catalogue_url = urljoin(environment.catalogue, identity_urls.identity_token)  # TODO would this be a new endpoint?
    sam_api_url = urljoin(environment.sam_api, sam_api_urls.sam_pat_token)

    sam_payload = {
        'username': user_id,
        'password': user_secret,
        'grant_type': 'password',

        # n.b. it makes no sense for us to have these per application, may as well be private once on Catalogue
        'client_id': os.environ["SAM_CLIENT_ID_FOR_PAT_EXCHANGE"],  # this must be for the one registered under 'Services' in PAT
        'client_secret': os.environ["SAM_CLIENT_SECRET_FOR_PAT_EXCHANGE"],  # this must be  for the one registered under 'Services' in PAT
    }
    sam_auth_header = {"apikey": os.environ["SAM_CLIENT_SECRET_FOR_PAT_EXCHANGE"]}

    logger.info("Exchanging details with SAM")
    resp = external_session().post(sam_api_url, data=sam_payload, headers=sam_auth_header)
    if resp.status_code != 200:
        raise Unauthorized('Could not retrieve jwt access token with client credentials')

    # exchange the acquired token with catalogue
    sam_token = resp.json()['access_token']
    dl_payload = {
        'client_id': "???", # TODO how would catalogue know its registered?
        'subject_token': sam_token
    }

    logger.info("Exchanging SAM token with Catalogue")
    resp = external_session().post(catalogue_url, data=dl_payload)
    if resp.status_code != 200:
        raise Unauthorized('Could not retrieve jwt access token with client credentials')

    jwt = resp.json()['access_token']
    return jwt


def decode_token(key):

    __pyjwt_algorithms = [
        # `HS256` is the value returned by the JWT from prod, but pyjwt seems
        # to complain about the signature.
        'HS256',
        'HS512', 'ES256', 'ES384', 'ES512', 'RS256', 'RS384',
        'RS512', 'PS256', 'PS384', 'PS512',
    ]

    return jwt.decode(
        jwt=key,
        algorithms=__pyjwt_algorithms,
        options={
            "verify_signature": False,
            "verify_aud": False,
            "verify_exp": False
        }
    )


def _pre_request_get_token(environment: _Environment, user_id: str, user_secret: str, hint: Optional[str]) -> JWTEncoded:

    _PAT = "PAT"
    _JWT = "JWT"
    _APP = "APP"

    if user_id is None or user_secret is None:
        raise Unauthorized()

    if not hint or hint == _PAT:
        # we think its a PAT or want to check it matches a UUID (which indicates it is)
        try:
            # is this a uuid as only PATs are UUIDs? - if it is it can be parse to UUID type
            uuid.UUID(user_id)
            return JWTEncoded(_sam_pat_flow(environment, user_id, user_secret))
        except (ValueError, TypeError) as e:
            logging.debug(f"{e} - doesn't look like a PAT {user_id}")
            if hint == _PAT:
                raise e

    if not hint or hint == _JWT:
        # we think its already a JWT or want to check its valid
        try:
            # is this a JWT? - if it is it can be decoded
            decode_token(user_id)
            return JWTEncoded(user_id)  # the JWT
        except jwt.DecodeError as e:
            logging.debug(f"{e} - doesn't look like a JWT {user_id}")
            if hint == _JWT:
                raise e

    if not hint or hint == _APP:
        # this seems to be an APP user, since its not a UUID or JWT
        try:
            return JWTEncoded(_sam_app_flow(environment, user_id, user_secret))
        except Exception as e:
            logging.debug(f"{e} - doesn't look like an APP {user_id}")
            if hint == _APP:
                raise e

    raise Unauthorized()



#
# Copyright (C) 2019 IHS Markit.
# All Rights Reserved
#
import datetime
import logging
import os
from collections import defaultdict, Mapping
from typing import Iterator, Optional, Iterable, Dict, List, NewType
from urllib.parse import urlparse
from dateutil.parser import isoparse
from flask import current_app

try:
    # these are only present for services, not for the SDK, and we do not wish to impose these in requirements.txt
    from injector import inject
except ModuleNotFoundError:
    # set up an Identity decorator for @inject
    def inject(function): # type: ignore
        def wrapper(*args, **kwargs):
            return function(*args, *kwargs)
        return wrapper

from ...adapters.s3_io_adapters import ServiceIO
from ...handlers.s3_dataset_handler import S3DatasetHandler
from ...handlers.s3_partition_handler import get_partitions_in_filepath, meets_partition_params
from ...services.exceptions import UnsupportedDatasetLocation, UnableToAccessDataset
from ...services.dlc_service import DlcService

logger = logging.getLogger(__name__)

# overwrite this in app.py with your choice of handler e.g.
# injector.binder.bind(DISPATCHER, to=ConsumptionDatasetHandler)
DISPATCHER = NewType('DISPATCHER', S3DatasetHandler)


class FileService:

    __HADOOP_FILES = {
        '_success'
    }

    @inject
    def __init__(
        self,
        dlc_service: DlcService
    ):
        self.dlc_service = dlc_service
        self.stream_handlers = {
            's3': current_app.injector.get(DISPATCHER),
        }
        logger.debug('FileService constructor')

    def __del__(self):
        """Destructor."""
        logger.debug('Destructor called on FileService')

    def get_handler(self, dataset):
        access_denied_message = (
            'You do not have access to this dataset. You must '
            'request access to the dataset\'s package.'
        )
        if 'location' not in dataset['attributes']:
            raise UnableToAccessDataset(details=access_denied_message)

        # If there is a 'location' key in the Catalogue response, then we expect it will contain
        # a type key.
        location_type: str = list(dataset['attributes']['location'].keys())[0].lower()

        if location_type == 'other':
            # Support old Catalogue behaviour where they return 'other' to mean access denied.
            raise UnableToAccessDataset(details=access_denied_message)

        # Retrieve the handler for this location type e.g. if the location is s3 then return the
        # s3 handler.
        if location_type in self.stream_handlers:
            return self.stream_handlers[location_type]
        else:
            supported_locations = ','.join(self.stream_handlers.keys()).lower()

            raise UnsupportedDatasetLocation(
                details=(
                    f'Only {supported_locations} currently '
                    'supports downloads via consumption. '
                    f'{location_type} is not currently supported'
                )
            )

    def get_files_from_datafile(
        self, datafile_id: str,
        paths: Optional[Iterable[str]] = None,
        partitions: Optional[Iterable] = None,
        order_by_latest: Optional[bool] = None,
        skip_hidden_files: Optional[bool] = False,
        skip_special_hadoop_files: Optional[bool] = False,
    ):
        """
        For a dataset_id returns a iterating response of Boto
        ObjectSummary objects. Can return a single file if
        path filter is an absolute path pointing at a specific file

        :param: datafile_id - the id of the datafile
        :kwarg: paths - a partial or complete path to a file or a
                              folder within a datafile
        :kwarg: partitions - a set of conditions to be met in the path
                     on comparison to each key=value pair related by key
                     e.g. `key` > 5 where key=4 in the object path is True
                     e.g. `key` < 4 where key=5 in the object path is False
        """
        logger.debug('get_files_from_datafile')
        datafile = self.dlc_service.get_datafile(datafile_id)
        dataset = self.dlc_service.get_dataset(
            datafile['properties']['datasetId']
        )

        return self._get_files(
            dataset=dataset,
            datafile=datafile,
            paths=paths,
            partitions=partitions,
            order_by_latest=order_by_latest,
            skip_hidden_files=skip_hidden_files,
            skip_special_hadoop_files=skip_special_hadoop_files,
        )

    # TODO merge with _get_files
    def get_files_from_dataset(
        self,
        dataset: Dict,
        paths: Optional[Iterable[str]] = None,
        partitions: Optional[Iterable] = None,
        order_by_latest: Optional[bool] = None,
        skip_hidden_files: Optional[bool] = False,
        skip_special_hadoop_files: Optional[bool] = False,
    ) -> Iterator[ServiceIO]:
        """
        For a dataset_id returns a iterating response of Boto
        ObjectSummary objects. Can return a single file if
        path filter is an absolute path pointing at a specific file

        :param: datafile_id - the id of the datafile
        :kwarg: paths - a partial or complete path to a file or a
                                 folder within a datafile
        :kwarg: partitions - a set of conditions to be met in the path
                         on comparison to each key=value pair related by key
                         e.g. `key` > 5 where key=4 in the object path is True
                         e.g. `key` < 4 where key=5 in the object path is False
        """
        logger.debug(
            'get_files_from_dataset',
        )

        yield from self._get_files(
            dataset=dataset,
            paths=paths,
            partitions=partitions,
            order_by_latest=order_by_latest,
            skip_hidden_files=skip_hidden_files,
            skip_special_hadoop_files=skip_special_hadoop_files,
        )

    def get_partitions_from_dataset(
        self,
        dataset: dict,
    ) -> Dict[str, list]:

        if not isinstance(dataset, Mapping):
            raise TypeError()
        logger.debug('get_partitions_from_dataset')
        partitions: Dict[str, list] = defaultdict(lambda: [])
        all_split = (
            get_partitions_in_filepath(f.path) for f in
            self.get_files_from_dataset(
                dataset,
            )
        )

        for file_path_splits in all_split:
            key_values = file_path_splits

            for k, v in key_values:
                partitions[k].append(v)

        for k in partitions.keys():
            partitions[k] = sorted(list(set(partitions[k])))

        return partitions

    @staticmethod
    def __filter_by_datafile(
        input_paths: Iterator[ServiceIO], datafile: dict
    ) -> Iterator[ServiceIO]:
        logger.debug('filter_by_datafile')
        filter_paths = [
            urlparse(f['path']).path.lstrip('/')
            for f in datafile['properties']['files']
        ]

        yield from FileService.__filter_by_paths(input_paths, filter_paths)

    # XXX Todo optimise this to be called first to filter paths
    # perhaps return some sort of q object?
    @staticmethod
    def __filter_by_paths(
        input_paths: Iterator[ServiceIO], filter_paths
    ) -> Iterator[ServiceIO]:
        logger.debug('filter_by_paths')
        for path in input_paths:
            if any([
                path.path.startswith(p) for p in filter_paths
            ]):
                yield path

    @staticmethod
    def __filter_by_partitions(
        input_paths: Iterator[ServiceIO], partitions
    ) -> Iterator[ServiceIO]:
        logger.debug('filter_by_partitions')
        for path in input_paths:
            if meets_partition_params(path.path, partitions):
                yield path

    def _get_files(
        self,
        dataset: Dict,
        datafile: Optional[Dict] = None,
        paths: Optional[Iterable[str]] = None,
        partitions: Optional[Iterable] = None,
        order_by_latest: Optional[bool] = None,
        skip_hidden_files: Optional[bool] = False,
        skip_special_hadoop_files: Optional[bool] = False,
    ) -> Iterator[ServiceIO]:
        logger.debug(
            '_get_files',
            extra={
                'order_by_latest': order_by_latest,
            }
        )

        handler = self.get_handler(dataset)

        # Common-prefixes level filtering
        result = handler.handle(
            dataset,
        )

        # Note: The order of operations below should be filter out files
        # before doing a sort because sort is expensive and some s3 listings
        # contain a lot of metadata (which is a waste to try to partition
        # or to even find the dates of the S3 files).

        # File-level filters
        if skip_hidden_files:
            logger.debug('skip_hidden_files')
            # [DL-4545][DL-4536] Do not read into files or directories that
            # are cruft from Spark which Spark will ignore on read,
            # e.g. files/dirs starting with `.` or `_` are hidden to Spark.

            def mis_hidden_file(path: str):
                attrs = dataset.get("attributes", None)
                if attrs:
                    content_type = attrs.get("content_type", "Structured")
                else:
                    content_type = "Structured"

                if (
                        path.startswith('.') or
                        path.startswith('_') or
                        (path == 'metadata' and content_type != "Unstructured") or
                        (
                            path.startswith('as_of_') and
                            path.endswith('=latest')
                        )
                ):
                    logger.debug(
                        f"skip_hidden_file: '{path}'",
                        extra={'path': path}
                    )
                    return True
                else:
                    return False

            result = (
                r for r in result
                if not any(mis_hidden_file(p) for p in r.path.split('/'))
            )

        if skip_special_hadoop_files:
            logger.debug('skip_special_hadoop_files')

            def is_special_hadoop_file(path: str):
                if os.path.basename(path).lower() in self.__HADOOP_FILES:
                    logger.warning(
                        f"skip_special_hadoop_file: '{path}'",
                        extra={'path': path}
                    )
                    return True
                else:
                    return False

            result = (
                r for r in result  # type: ignore
                if not is_special_hadoop_file(r.path)
            )

        if datafile:
            # When we are passed a datafile, we only get the files specified
            # in the datafile.
            result = self.__filter_by_datafile(result, datafile)

        if paths:
            result = self.__filter_by_paths(result, paths)

        if partitions:
            result = self.__filter_by_partitions(result, partitions)

        if order_by_latest is None:
            pass
        elif order_by_latest is True:
            result = FileService.order_by_latest(result, direction='desc')
        elif order_by_latest is False:
            result = FileService.order_by_latest(result, direction='asc')

        yield from result

    @staticmethod
    def order_by_latest(input_paths: Iterator[ServiceIO], direction='desc'
                        ) -> Iterator[ServiceIO]:
        reverse = True if direction == 'desc' else False

        logger.debug(
            'order_by_latest',
            extra={
                'direction': direction,
                'reverse': reverse,
            }
        )

        if reverse:
            # Use list so we do not break iterator when we sort.
            input_paths_list: List[ServiceIO] = list(input_paths)
            # Sort in-place to avoid allocating more objects in Python free-list.
            input_paths_list.sort(
                key=FileService._get_created_at,
                reverse=reverse
            )

            return iter(input_paths_list)
        else:
            return input_paths

    @staticmethod
    def _get_created_at(service_io_obj: ServiceIO):
        partitions = {
            k.lower(): v for k, v in dict(
                get_partitions_in_filepath(service_io_obj.path)
            ).items()
        }

        chronological_partitions = {
            'year', 'month', 'day'
        }

        if partitions:
            as_of = next(
                (v for k, v in partitions.items()
                 if k.startswith('as_of_')),
                None
            )

            if as_of:
                try:
                    as_of = isoparse(as_of)
                    return (
                        3,
                        as_of,
                        service_io_obj.path
                    )
                except ValueError:
                    pass

            # If no as_of key, fallback on chronological_partitions
            if any(
                i in chronological_partitions for i in partitions.keys()
            ):
                year = int(partitions.get('year', 1))
                month = int(partitions.get('month', 1))
                day = int(partitions.get('day', 1))

                return (
                    2,
                    datetime.datetime(year, month, day),
                    service_io_obj.path
                )

        # If no partitions order by S3 dateting
        if 'last_modified' in service_io_obj.metadata:
            # Secondary sort to rely on the object's S3 metadata
            # if partitions can not be found
            return (
                1,
                service_io_obj.metadata['last_modified'],
                service_io_obj.path
            )
        else:
            # order by path and hope for the best
            return (
                0,
                # Dummy date
                datetime.datetime.min,
                service_io_obj.path
            )


import types
from functools import wraps

from . import _make_function_call
from .analytics_aspect import AnalyticsAspect
from .logging_aspect import LoggingAspect


class ComponentsAspectWrapper(type):
    """
    This decorates all functions in a Component with a logging function, the equivalent to decorators::log_public_functions_calls_using()
    This class should be used to extend the type i.e. A(ComponentsAspectWrapper), applying __aspects to all its methods*
    """
    __aspects = [LoggingAspect(), AnalyticsAspect()]

    def __new__(cls, name, bases, attrs):
        for attr_name, attr_value in attrs.items():
            if (
                isinstance(attr_value, types.FunctionType)
                and not attr_name.startswith('_')
            ):
                # add the decorator to cls.attr_value if attr_value is a function
                attrs[attr_name] = cls._wrap_call_with_aspects(attr_value)

        return super(ComponentsAspectWrapper, cls).__new__(cls, name, bases, attrs)

    @classmethod
    def _wrap_call_with_aspects(cls, func):
        @wraps(func)
        def wrapper(self, *args, **kwargs):
            return _make_function_call(cls.__aspects, self, func, args, kwargs, class_fields_to_include=None)

        return wrapper
